__all__ = ('CooldownHandler', )

from ...backend.event_loop import LOOP_TIME

from ...discord.channel import ChannelGuildBase
from ...discord.core import KOKORO

from .exceptions import CommandCooldownError

class CooldownUnit:
    """
    A cooldown unit stored by a ``CommandCooldownWrapper``-s.
    
    Attributes
    ----------
    expires_at : `float`
        When the cooldown unit will expire in LOOP_TIME time.
    uses_left : `int`
        How much uses are left till the respective entity will be locked by cooldown.
    """
    __slots__ = ('expires_at', 'uses_left',)
    def __init__(self, expires_at, uses_left):
        """
        Creates a new ``CooldownUnit`` with the given parameters.
        
        Parameters
        ----------
        expires_at : `float`
            When the cooldown unit will expire in LOOP_TIME time.
        uses_left : `int`
            How much uses are left till the respective entity will be locked by cooldown.
        """
        self.expires_at = expires_at
        self.uses_left = uses_left
    
    def __repr__(self):
        """Returns the object's representation."""
        return f'{self.__class__.__name__}(expires_at={self.expires_at}, uses_left={self.uses_left})'


def _check_user(cooldown_handler, command_context):
    """
    Executes user cooldown check.
    
    Might be set as the ``Cooldown``'s ``.checker`` instance attribute.
    
    Parameters
    ----------
    cooldown_handler : ``CooldownHandler``
        The parent cooldown handler.
    command_context : ``CommandHandler``
        The received command's context.
    
    Returns
    -------
    expires_at : `int`
        When the cooldown for the given entity will expire.
    """
    user_id = command_context.message.author.id
    
    cache = cooldown_handler.cache
    try:
        unit = cache[user_id]
    except KeyError:
        at_ = LOOP_TIME()+cooldown_handler.reset
        cache[user_id] = CooldownUnit(at_, cooldown_handler.limit)
        KOKORO.call_at(at_, dict.__delitem__, cache, user_id)
        return 0.
    
    left = unit.uses_left
    if left > 0:
        unit.uses_left = left-cooldown_handler.weight
        return 0.
    
    return unit.expires_at


def _check_channel(cooldown_handler, command_context):
    """
    Executes channel cooldown check.
    
    Might be set as the ``Cooldown``'s ``.checker`` instance attribute.
    
    Parameters
    ----------
    cooldown_handler : ``CooldownHandler``
        The parent cooldown handler.
    command_context : ``CommandHandler``
        The received command's context.
    
    Returns
    -------
    expires_at : `int`
        When the cooldown for the given entity will expire.
    """
    channel_id =command_context. message.channel.id
    
    cache = cooldown_handler.cache
    try:
        unit = cache[channel_id]
    except KeyError:
        at_ = LOOP_TIME()+cooldown_handler.reset
        cache[channel_id] = CooldownUnit(at_, cooldown_handler.limit)
        KOKORO.call_at(at_, dict.__delitem__, cache, channel_id)
        return 0.
    
    left = unit.uses_left
    if left > 0:
        unit.uses_left = left-cooldown_handler.weight
        return 0.
    
    return unit.expires_at


def _check_guild(cooldown_handler, command_context):
    """
    Executes guild based cooldown check.
    
    Might be set as the ``Cooldown``'s ``.checker`` instance attribute.
    
    Parameters
    ----------
    cooldown_handler : ``CooldownHandler``
        The parent cooldown handler.
    command_context : ``CommandHandler``
        The received command's context.
    
    Returns
    -------
    expires_at : `int`
        When the cooldown for the given entity will expire.
        
        If the cooldown limitation is not applicable for the given entity, returns `-1.0`.
    """
    channel = command_context.message.channel
    if not isinstance(channel, ChannelGuildBase):
        return -1.0
    
    guild_id = channel.guild.id
    
    cache = cooldown_handler.cache
    try:
        unit = cache[guild_id]
    except KeyError:
        at_ = LOOP_TIME()+cooldown_handler.reset
        cache[guild_id] = CooldownUnit(at_, command_context.self.limit)
        KOKORO.call_at(at_, dict.__delitem__, cache, guild_id)
        return 0.
    
    left = unit.uses_left
    if left > 0:
        unit.uses_left = left-cooldown_handler.weight
        return 0.
    
    return unit.expires_at


class CooldownHandler:
    """
    Cooldown for commands.
    
    Attributes
    ----------
    cache : `dict` of (``DiscordEntity``, ``CooldownUnit``) items
        Cache to remember how much use of the given entity are exhausted already.
    checker : `function`
        Checks after how much time the given entity can use again the respective command.
    limit : `int`
        The amount of how much times the command can be called within a set duration before going on cooldown.
    reset : `float`
        The time after the cooldown resets.
    weight : `int`
        The weight of the command.
    """
    __slots__ = ('cache', 'checker', 'limit', 'reset', 'weight',)
    
    def __new__(cls, for_, reset, limit=1, weight=1):
        """
        Creates a new ``CooldownHandler`` instance from the given parameters.
        
        Parameters
        ----------
        for_ : `str`
            By what type of entity the cooldown should limit the command.
            
            Possible values:
             - `'user'`
             - `'channel'`
             - `'guild'`
         
        reset : `float`
            The reset time of the cooldown.
        
        limit : `int`
            The amount of calls after the respective command goes on cooldown.
        
        weight : `int`, Optional
            The weight of one call. Defaults to `1`.
        
        Raises
        ------
        TypeError
            - If `str` is not given as `str` instance.
            - If `weight` is not numeric convertable to `int`.
            - If `reset` is not numeric convertable to `float`.
            - If `limit` is not numeric convertable to `int`.
        ValueError
            - If `for_` is not given as any of the expected value.
        """
        for_type = type(for_)
        if for_type is str:
            pass
        elif issubclass(for_, str):
            for_ = str(for_)
        else:
            raise TypeError(f'`for_` can be given as `str` instance, got {for_type.__name__}.')
        
        if 'user'.startswith(for_):
            checker = _check_user
        elif 'channel'.startswith(for_):
            checker = _check_channel
        elif 'guild'.startswith(for_):
            checker = _check_guild
        else:
            raise ValueError(f'\'for_\' can be \'user\', \'channel\' or \'guild\', got {for_!r}')
        
        reset_type = reset.__class__
        if (reset_type is not float):
            try:
                __float__ = getattr(reset_type, '__float__')
            except AttributeError:
                raise TypeError(f'The given reset is not `float`, neither other numeric convertable to it, got '
                    f'{reset_type.__name__}.') from None
            
            reset = __float__(reset)
            
        limit_type = limit.__class__
        if limit_type is int:
            pass
        elif issubclass(limit_type, int):
            limit = int(limit)
        else:
            raise TypeError(f'`limit` can be given as `int` instance, got {limit_type.__name__}.') from None
        
        weight_type = weight.__class__
        if weight_type is int:
            pass
        elif issubclass(weight_type, int):
            weight = int(weight)
        else:
            raise TypeError(f'`weight` can be given as `int` instance, got {weight_type.__name__}.') from None
        
        self = object.__new__(cls)
        self.checker = checker
        self.reset = reset
        self.weight = weight
        self.limit = limit-weight
        self.cache = {}
        
        return self
    
    
    async def __call__(self, command_context):
        """
        Calls the cooldown with the respective `client` and `message`, and then yields whether the command can be
        called, and if not, then with what extra parameters the handler should receive.
        
        This method is a coroutine.
        
        Parameters
        ----------
        command_context : ``CommandHandler``
            The received command's context.
        
        Raises
        ------
        CommandCooldownError
            If the command is on cooldown, or if guild-bound cooldown was called from non-guild.
        """
        expires_at = self.checker(self, command_context)
        if expires_at:
            if expires_at == -1:
                expires_after = -1
            else:
                expires_after = expires_at-LOOP_TIME()
            
            raise CommandCooldownError(self, expires_after)



__all__ = ('Sticker', )

import warnings

from ...backend.export import include

from ..core import STICKERS, GUILDS, STICKER_PACKS
from ..bases import DiscordEntity
from ..user import ZEROUSER, User
from ..http import urls as module_urls
from ..preconverters import preconvert_str, preconvert_int, preconvert_snowflake, preconvert_preinstanced_type, \
    preconvert_iterable_of_str, preconvert_bool
from ..bases import instance_or_id_to_instance

from .preinstanced import StickerFormat, StickerType

Client = include('Client')

DEFAULT_SORT_VALUE = 0

class Sticker(DiscordEntity, immortal=True):
    """
    Represents a ``Message``'s sticker.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the sticker.
    available : `bool`
        Whether the sticker is available.
    description : `None` or `str`
        The sticker's description.
    format : ``StickerFormat``
        The sticker's format.
    guild_id : `int`
        The guild's identifier to what the sticker is bound to. Defaults to `0` if the sticker is not bound to any
        guild.
    name : `str`
        The sticker's name.
    pack_id : `int`
        The unique identifier number of the pack from the sticker is. Defaults to `0` if the sticker is not bound to
        any pack.
    sort_value : `int`
        Value used to sort the stickers.
    tags : None` or `frozenset` of `str`
        Tags of the sticker if applicable.
    type : ``StickerType``
        The sticker's type.
    user : ``ClientUserBase``
        The user who uploaded the emoji. Defaults to ``ZEROUSER``.
    """
    __slots__ = ('available', 'description', 'format', 'guild_id', 'name', 'pack_id', 'sort_value', 'tags',
        'type', 'user')
    
    def __new__(cls, data):
        """
        Creates a new ``Sticker`` from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Sticker data.
        """
        sticker_id = int(data['id'])
        
        try:
            self = STICKERS[sticker_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = sticker_id
            STICKERS[sticker_id] = self
            
            try:
                user_data = data['user']
            except KeyError:
                user = ZEROUSER
            else:
                user = User(user_data)
            self.user = user
        
        else:
            if not self.partial:
                if self.user is ZEROUSER:
                    try:
                        user_data = data['user']
                    except KeyError:
                        pass
                    else:
                        self.user = User(user_data)
                
                return self
        
        self._update_from_partial(data)
        
        return self
    
    
    @classmethod
    def _create_partial(cls, data):
        """
        Creates a sticker from the given partial ``Sticker`` data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Sticker data.
        
        Returns
        -------
        self : ``Sticker``
        """
        sticker_id = int(data['id'])
        
        try:
            self = STICKERS[sticker_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = sticker_id
            STICKERS[sticker_id] = self
            
            self.available = True
            self.description = None
            self.guild_id = 0
            self.pack_id = 0
            self.sort_value = DEFAULT_SORT_VALUE
            self.tags = None
            self.type = StickerType.none
            self.user = ZEROUSER
        else:
            if not self.partial:
                return self
        
        self.format = StickerFormat.get(data.get('format_type', 0))
        self.name = data['name']
        
        return self
    
    def _update_from_partial(self, data):
        """
        Updates a partial sticker with to not partial from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Sticker data.
        """
        self.format = StickerFormat.get(data.get('format_type', 0))
        
        pack_id = data.get('pack_id', None)
        if pack_id is None:
            pack_id = 0
        else:
            pack_id = int(pack_id)
        self.pack_id = pack_id
        
        guild_id = data.get('guild_id', None)
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        self.guild_id = guild_id
        
        self.type = StickerType.get(data.get('type', 0))
        
        self._update_attributes(data)
    
    
    def _update_attributes(self, data):
        """
        Updates the stickers with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Sticker data.
        """
        self.description = data.get('description', None)
        self.name = data['name']
        self.sort_value = data.get('sort_value', DEFAULT_SORT_VALUE)
        
        tags = data.get('tags', None)
        if (tags is None) or (not tags):
            tags = None
        else:
            tags = frozenset(tags.split(', '))
        self.tags = tags
        
        self.available = data.get('available', True)
    
    def _difference_update_attributes(self, data):
        """
        Updates the sticker with the given data and returns the changed attributes in `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Sticker data.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dictionary is optional.
            
            +-----------------------+-----------------------------------+
            | Keys                  | Values                            |
            +=======================+===================================+
            | available             | `bool`                            |
            +-----------------------+-----------------------------------+
            | description           | `None` or `str`                   |
            +-----------------------+-----------------------------------+
            | name                  | `str`                             |
            +-----------------------+-----------------------------------+
            | sort_value            | `int`                             |
            +-----------------------+-----------------------------------+
            | tags                  | `None`  or `frozenset` of `str`   |
            +-----------------------+-----------------------------------+
        """
        old_attributes = {}
        
        description = data.get('description', None)
        if description != self.description:
            old_attributes['description'] = self.description
            self.description = description
        
        name = data['name']
        if name != self.name:
            old_attributes['name'] = self.name
            self.name = name
        
        sort_value = data.get('sort_value', DEFAULT_SORT_VALUE)
        if sort_value != self.sort_value:
            old_attributes['sort_value'] = self.sort_value
            self.sort_value = sort_value
        
        tags = data.get('tags', None)
        if (tags is None) or (not tags):
            tags = None
        else:
            tags = frozenset(tags.split(', '))
        if tags != self.tags:
            old_attributes['tags'] = self.tags
            self.tags = tags
        
        available = data.get('available', True)
        if self.available != available:
            old_attributes['available'] = self.available
            self.available = available
        
        return old_attributes
    
    
    def __repr__(self):
        """Returns the sticker's representation."""
        return f'<{self.__class__.__name__} id={self.id!r}, name={self.name!r}>'
    
    url = property(module_urls.sticker_url)
    url_as = module_urls.sticker_url_as
    
    def __gt__(self, other):
        """Returns whether self is greater than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_sort_value = self.sort_value
        other_sort_value = other.sort_value
        if self_sort_value > other_sort_value:
            return True
        
        if self_sort_value == other_sort_value:
            if self.id > other.id:
                return True
        
        return False
    
    def __ge__(self, other):
        """Returns whether self is greater or equal to other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_sort_value = self.sort_value
        other_sort_value = other.sort_value
        if self_sort_value > other_sort_value:
            return True
        
        if self_sort_value == other_sort_value:
            if self.id >= other.id:
                return True
        
        return False
    
    def __le__(self, other):
        """Returns whether self is less or equal to other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_sort_value = self.sort_value
        other_sort_value = other.sort_value
        if self_sort_value < other_sort_value:
            return True
        
        if self_sort_value == other_sort_value:
            if self.id <= other.id:
                return True
        
        return False
    
    def __lt__(self, other):
        """Returns whether self is less than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_sort_value = self.sort_value
        other_sort_value = other.sort_value
        if self_sort_value < other_sort_value:
            return True
        
        if self_sort_value == other_sort_value:
            if self.id < other.id:
                return True
        
        return False
    
    @property
    def partial(self):
        """
        Returns whether the sticker is partial.
        
        Returns
        -------
        partial : `bool`
        """
        pack_id = self.pack_id
        if pack_id:
            if pack_id in STICKER_PACKS:
                return False
            
            return True
        
        guild_id = self.guild_id
        if guild_id:
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                return False
            else:
                return guild.partial
        
        return True
    
    
    @classmethod
    def precreate(cls, sticker_id, **kwargs):
        """
        Precreates the sticker by creating a partial one with the given parameters. When the sticker is loaded
        the precrated one will be picked up. If an already existing sticker would be precreated, returns that
        instead and updates that only, if that is partial.
        
        Parameters
        ----------
        sticker_id : `snowflake`
            The sticker's id.
        **kwargs : keyword parameters
            Additional predefined attributes for the sticker.
        
        Other Parameters
        ----------------
        available : `bool`, Optional (Keyword only)
             The sticker's ``.available``.
        description : `str`, Optional (Keyword only)
            The sticker's ``.name``. It's length can be in range [0:1024]
        format : ``StickerFormat``, Optional (Keyword only)
            The sticker's ``.format``.
        guild_id : `int`, Optional (Keyword only)
            The sticker's ``.guild_id``.
        name : `str`, Optional (Keyword only)
            The sticker's ``.name``. It's length can be in range [0:32].
        pack_id : `int`, Optional (Keyword only)
            The sticker's ``.pack_id``.
        sort_value : `int`, Optional (Keyword only)
            The sticker's ``.sort_value``.
        tags : `None` or `iterable` of `str`, Optional (Keyword only)
            The sticker's ``.tags``.
        type : ``StickerType``, Optional (Keyword only)
            The sticker's ``.type``.
        user : ``ClientUserBase``, Optional (Keyword only)
            The sticker's ``.user``.
        
        Returns
        -------
        self : ``Sticker``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        sticker_id = preconvert_snowflake(sticker_id, 'sticker_id')
        
        if kwargs:
            processable = []
            # There might be new bool fields coming out.
            for attribute_name in ('available',):
                try:
                    attribute_value = kwargs.pop(attribute_name)
                except KeyError:
                    pass
                else:
                    animated = preconvert_bool(attribute_value, attribute_name)
                    processable.append((attribute_name, animated))
            
            for attribute_name, lower_limit, upper_limit in (
                        ('description', 0, 1024),
                        ('name', 2, 32),
                    ):
                
                try:
                    attribute_value = kwargs.pop('attribute_name')
                except KeyError:
                    pass
                else:
                    attribute_value = preconvert_str(attribute_value, attribute_name, lower_limit, upper_limit)
                    processable.append((attribute_name, attribute_value))
            
            for attribute_name, preinstanced_type in (
                        ('format_type', StickerFormat),
                        ('type', StickerType),
                    ):
                try:
                    attribute_value = kwargs.pop('attribute_name')
                except KeyError:
                    pass
                else:
                    attribute_value = preconvert_preinstanced_type(attribute_value, attribute_name, preinstanced_type)
                    processable.append((attribute_name, attribute_value))
            
            for attribute_name in ('guild_id', 'pack_id'):
                try:
                    attribute_value = kwargs.pop(attribute_name)
                except KeyError:
                    pass
                else:
                    attribute_value = preconvert_snowflake(attribute_value, attribute_name)
                    processable.append((attribute_name, attribute_value))
            
            try:
                sort_value = kwargs.pop('sort_value')
            except KeyError:
                pass
            else:
                sort_value = preconvert_int(sort_value, 'sort_value', 0, (1<<16)-1)
                processable.append(('sort_value', sort_value))
            
            try:
                tags = kwargs.pop('tags')
            except KeyError:
                pass
            else:
                if (tags is not None):
                    tags = preconvert_iterable_of_str(tags, 'tags', 0, 256, 2, 32)
                    if tags:
                        tags = frozenset(tags)
                    else:
                        tags = None
                
                processable.append(('tags', tags))
            
            try:
                user = kwargs.pop('user')
            except KeyError:
                pass
            else:
                user = instance_or_id_to_instance(user, (User, Client), 'user')
                processable.append(('user', user))
            
            
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs}')
        
        else:
            processable = None
        
        
        try:
            self = STICKERS[sticker_id]
        except KeyError:
            self = cls._create_empty(sticker_id)
            STICKERS[sticker_id] = self
        else:
            if not self.partial:
                return self
        
        if (processable is not None):
            for name, value in processable:
                setattr(self, name, value)
        
        return self
    
    
    @classmethod
    def _create_empty(cls, sticker_id):
        """
        Creates an empty sticker with the given identifier.
        
        Parameters
        ----------
        sticker_id : `int`
            The sticker's identifier.
        
        Returns
        -------
        self : ``Sticker``
            The created sticker.
        """
        self = object.__new__(cls)
        self.id = sticker_id
        self.available = True
        self.description = ''
        self.format = StickerFormat.none
        self.guild_id = 0
        self.name = ''
        self.pack_id = 0
        self.sort_value = DEFAULT_SORT_VALUE
        self.tags = None
        self.type = StickerType.none
        self.user = ZEROUSER
        return self
    
    
    @property
    def guild(self):
        """
        Returns the sticker's guild if cached.
        
        Returns
        -------
        guild : ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)



__all__ = ()

from ...backend.futures import is_coroutine_generator

from ...discord.exceptions import DiscordException, ERROR_CODES
from ...discord.embed import EmbedBase

def is_only_embed(maybe_embeds):
    """
    Checks whether the given value is a `tuple` or `list` containing only `embed-like`-s.
    
    Parameters
    ----------
    maybe_embeds : (`tuple` or `list`) of `EmbedBase` or `Any`
        The value to check whether is a `tuple` or `list` containing only `embed-like`-s.
    
    Returns
    -------
    is_only_embed : `bool`
    """
    if not isinstance(maybe_embeds, (list, tuple)):
        return False
    
    for maybe_embed in maybe_embeds:
        if not isinstance(maybe_embed, EmbedBase):
            return False
    
    return True


async def send_response(command_context, response):
    """
    Sends to discord the given response.
    
    This function is a coroutine.
    
    Parameters
    ----------
    command_context : ``CommandContext``
        The respective command context.
    response : `Any`
        Any object yielded or returned by the command coroutine.
    
    Returns
    -------
    message : `None` or ``Message``
        The sent message if any.
    """
    if (response is None):
        return
    
    if isinstance(response, str):
        return await command_context.client.message_create(command_context.channel, response)
        
    if isinstance(response, EmbedBase) or is_only_embed(response) and response:
        return await command_context.client.message_create(command_context.channel, embed=response)
    
    if is_coroutine_generator(response):
        return await process_command_coroutine_generator(command_context, response)
    
    response = str(response)
    if len(response) > 2000:
        response = response[:2000]
    
    if response:
        return await command_context.client.message_create(command_context.channel, response)
    
    # No more cases
    return


async def process_command_coroutine_generator(command_context, coroutine_generator):
    """
    Processes a command coroutine generator.
    
    This function is a coroutine.
    
    Parameters
    ----------
    command_context : ``CommandContext``
        The respective command context.
    coroutine_generator : `CoroutineGenerator`
        A coroutine generator with will send command response.
    
    Returns
    -------
    response : `Any`
        Returned object by the coroutine generator.
    
    Raises
    ------
    BaseException
        Any exception raised by `coroutine_generator`.
    """
    response_message = None
    response_exception = None
    while True:
        if response_exception is None:
            step = coroutine_generator.asend(response_message)
        else:
            step = coroutine_generator.athrow(response_exception)
        
        try:
            response = await step
        except StopAsyncIteration as err:
            # catch `StopAsyncIteration` only if it is a new one.
            if (response_exception is not None) and (response_exception is not err):
                raise
            
            args = err.args
            if args:
                response = args[0]
            else:
                response = None
            break
        
        except BaseException as err:
            if (response_exception is None) or (response_exception is not err):
                raise
            
            if isinstance(err, ConnectionError):
                return
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_channel, # Message's channel deleted; Can we get this?
                        ERROR_CODES.missing_access, # Client removed.
                            ):
                    return
            
            raise
        
        else:
            try:
                response_message = await send_response(command_context, response)
            except BaseException as err:
                response_message = None
                response_exception = err
            else:
                response_exception = None
    
    return response


async def process_command_coroutine(command_context, coroutine):
    """
    Processes a command coroutine.
    
    If the coroutine returns or yields a string or an embed like then sends it to the respective channel.
    
    This function is a coroutine.
    
    Parameters
    ----------
    command_context : ``CommandContext``
        The respective command context.
    coroutine : `Coroutine`
        A coroutine with will send command response.
    
    Raises
    ------
    BaseException
        Any exception raised by `coro`.
    """
    if is_coroutine_generator(coroutine):
        response = await process_command_coroutine_generator(command_context, coroutine)
    else:
        response = await coroutine
    
    await send_response(command_context, response)



# -*- coding: utf-8 -*-
"""
Checks can be added to commands or to categories to limit their usage to set users or places.

The implemented checks are the following:

+--------------------------------+-----------------+--------------------------------------------------------------+
| Name                           | Extra parameter | Description                                                  |
+================================+=================+==============================================================+
| announcement_channel_only      | N/A             | Whether the message's channel is an announcement channel.    |
+--------------------------------+-----------------+--------------------------------------------------------------+
| booster_only                   | N/A             | Whether the user boosts the respective guild.                |
+--------------------------------+-----------------+--------------------------------------------------------------+
| bot_account_only               | N/A             | Whether the message's author is a bot account.               |
+--------------------------------+-----------------+--------------------------------------------------------------+
| client_has_guild_permissions   | permissions     | Whether the client has the given permissions at the guild.   |
|                                |                 | (Fails in private channels.)                                 |
+--------------------------------+-----------------+--------------------------------------------------------------+
| client_has_permissions         | permissions     | Whether the client has the given permissions at the channel. |
+--------------------------------+-----------------+--------------------------------------------------------------+
| client_only                    | N/A             | Whether the message was sent by a ``Client``.                |
+--------------------------------+-----------------+--------------------------------------------------------------+
| custom                         | function        | Custom checks, to wrap a given `function`. (Can be async.)   |
+--------------------------------+-----------------+--------------------------------------------------------------+
| guild_only                     | N/A             | Whether the message was sent to a guild channel.             |
+--------------------------------+-----------------+--------------------------------------------------------------+
| guild_owner                    | N/A             | Whether the message's author is the guild's owner.           |
|                                |                 | (Fails in private channels.)                                 |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_any_role                   | roles           | Whether the message's author has any of the given roles.     |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_guild_permissions          | permissions     | Whether the message's author has the given permissions at    |
|                                |                 | the guild. (Fails in private channels.)                      |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_permissions                | permissions     | Whether the message's author has the given permissions at    |
|                                |                 | the channel.                                                 |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_role                       | role            | Whether the message's author has the given role.             |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_any_channel                 | channels        | Whether the message was sent to any of the given channels.   |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_any_guild                   | guilds          | Whether the message was sent to any of the given guilds.     |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_channel                     | channel         | Whether the message's channel is the given one.              |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_guild                       | guild           | Whether the message guild is the given one.                  |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_in_any_category             | categories      | Whether the message was sent into a channel, what's category |
|                                |                 | is any of the specified ones.                                |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_in_category                 | category        | Whether the message was sent into a channel, what's category |
|                                |                 | is the specified one.                                        |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_in_voice                    | N/A             | Whether the user is in a voice channel in the respective     |
|                                |                 | guild.                                                       |
+--------------------------------+-----------------+--------------------------------------------------------------+
| nsfw_channel_only              | N/A             | Whether the message's channel is nsfw.                       |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_only                     | N/A             | Whether the message's author is an owner of the client.      |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_guild_owner           | N/A             | `owner_only` or `guild_owner` (Fails in private channels.)   |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_has_any_role          | roles           | `owner_only` or `has_any_role`                               |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_has_guild_permissions | permissions     | `owner_only` or `has_guild_permissions`                      |
|                                |                 | (Fails in private channels.)                                 |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_has_permissions       | permissions     | `owner_only` or `has_permissions`                            |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_has_role              | role            | `owner_only` or `has_any_role`                               |
+--------------------------------+-----------------+--------------------------------------------------------------+
| private_only                   | N/A             | Whether the message's channel is a private channel.          |
+--------------------------------+-----------------+--------------------------------------------------------------+
| user_account_only              | N/A             | Whether the message's author is a user account.              |
+--------------------------------+-----------------+--------------------------------------------------------------+
| user_account_or_client         | N/A             | Whether the message's author is a user account or a          |
|                                |                 | ``Client`` instance.                                         |
+--------------------------------+-----------------+--------------------------------------------------------------+

Every check also accepts an additional keyword parameter, called `handler`, what is called, when the respective
check fails (returns `False`).

To a check's handler the following parameters are passed:

+-------------------+---------------------------+
| Respective name   | Type                      |
+===================+===========================+
| client            | ``Client``                |
+-------------------+---------------------------+
| message           | ``Message``               |
+-------------------+---------------------------+
| command           | ``Command`` or `str`      |
+-------------------+---------------------------+
| check             | ``_check_base`` instance  |
+-------------------+---------------------------+

If a command's check fails, then `command` is given as `Command` instance, tho checks can be added not only to
commands and at those cases, `command` is given as `str`.
"""

from ...backend.utils import NEEDS_DUMMY_INIT
from ...backend.futures import Task
from ...backend.analyzer import CallableAnalyzer

from ...discord.core import KOKORO
from ...discord.bases import instance_or_id_to_instance, instance_or_id_to_snowflake
from ...discord.guild import Guild
from ...discord.permission import Permission
from ...discord.role import Role
from ...discord.channel import ChannelBase, ChannelText, ChannelCategory, ChannelGuildBase
from ...discord.events.handling_helpers import check_parameter_count_and_convert
from ...discord.client import Client

def validate_checks(checks_):
    """
    Validates the given checks.
    
    Parameters
    ----------
    checks_ : `None`, ``_check_base`` instance or `list` of ``_check_base`` instances
        Checks to define in which circumstances a command should be called.
        
    Returns
    -------
    checks_processed : `None` or `list` of ``_check_base``
        Will never return an empty list.
    
    Raises
    ------
    TypeError
        If `checks_` was not given as `None` or as `list` of ``_check_base`` instances.
    """
    if checks_ is None:
        checks_processed = None
    elif isinstance(checks_, _check_base):
        checks_processed = [checks_]
    elif isinstance(checks_, list):
        checks_processed = []
        
        for index, check in enumerate(checks_):
            check_type = check.__class__
            if issubclass(check_type, _check_base):
                checks_processed.append(check)
                continue
            
            raise TypeError(f'`checks` element {index} was not given as `{_check_base.__name__}`, got '
                f'`{check_type.__name__}`.')
        
        if not checks_processed:
            checks_processed = None
    else:
        raise TypeError(f'`checks_` should have been given as `None`, `{_check_base.__name__}` instance or as '
            f'`list` of `{_check_base.__name__}` instances, got: {checks_.__class__.__name__}; {checks_!r}.')
    
    # Unwrap `and` checks
    if (checks_processed is not None):
        index = 0
        limit = len(checks_processed)
        while True:
            check = checks_processed[index]
            if isinstance(check, _and_op_check):
                del checks_processed[index]
                for check in check.checks:
                    checks_processed.append(check)
                
                limit = len(checks_processed)
            else:
                index += 1
            
            if index >= limit:
                break
    
    return checks_processed


def _convert_handler(handler):
    """
    Validates the given handler.
    
    Parameters
    ----------
    handler : `None` or `async-callable` or instantiable to `async-callable`
        The handler to convert.
        
        If the handler is `async-callable` or if it would be instanced to it, then it should accept the following
        parameters:
        +-------------------+---------------------------+
        | Respective name   | Type                      |
        +===================+===========================+
        | client            | ``Client``                |
        +-------------------+---------------------------+
        | message           | ``Message``               |
        +-------------------+---------------------------+
        | command           | ``Command`` or `str`      |
        +-------------------+---------------------------+
        | check             | ``_check_base`` instance  |
        +-------------------+---------------------------+
    
    Returns
    -------
    handler : `None` or `async-callable`
    
    Raises
    ------
    TypeError
        If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
    """
    if (handler is not None):
        handler = check_parameter_count_and_convert(handler, 4, name='handler', error_message= \
            '`handler` expects to pass 4 parameters (client, message, command, check).')
    return handler

def _convert_permissions(permissions):
    """
    Validates the given `permissions`.
    
    Parameters
    ----------
    permissions : ``Permission`` or `int` instance
        Permission to validate.
    
    Returns
    -------
    permissions : ``Permission``
    
    Raises
    ------
    TypeError
        `permissions` was not given as `int` instance.
    """
    permission_type = permissions.__class__
    if permission_type is Permission:
        pass
    elif issubclass(permission_type, int):
        permissions = Permission(permissions)
    else:
        raise TypeError(f'`permissions` should have been passed as a `{Permission.__name__}` object or as an '
            f'`int` instance, got {permission_type.__name__}.')
    
    return permissions

class _check_meta(type):
    """
    Check metaclass for collecting their slots in a `SLOTS` class attribute.
    """
    def __new__(cls, class_name, class_parents, class_attributes):
        """
        Parameters
        ----------
        class_name : `str`
            The created class's name.
        class_parents : `tuple` of `type` instances
            The superclasses of the creates type.
        class_attributes : `dict` of (`str`, `Any`) items
            The class attributes of the created type.
        
        Returns
        -------
        type : ``_check_meta`` instance
        """
        if class_parents:
            parent = class_parents[0]
            inherited_slots = getattr(parent, 'SLOTS', None)
        else:
            inherited_slots = None
        
        new_slots = class_attributes.get('__slots__', None)
        
        final_slots = []
        if (inherited_slots is not None):
            final_slots.extend(inherited_slots)
        
        if (new_slots is not None):
            final_slots.extend(new_slots)
        
        class_attributes['SLOTS'] = tuple(final_slots)
        
        return type.__new__(cls, class_name, class_parents, class_attributes)

class _check_base(metaclass=_check_meta):
    """
    Base class for checks.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ('handler',)
    
    def __new__(cls, handler=None):
        """
        Creates a check with the given parameters.
        
        Parameters
        ----------
        handler : `None` or `async-callable` or instantiable to `async-callable`
            Will be called when the check fails.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        """
        handler = _convert_handler(handler)
        self = object.__new__(cls)
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        Subclasses should overwrite this method.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        return False
    
    def __repr__(self):
        """Returns the check's representation."""
        result = [
            self.__class__.__name__,
            '(',
                ]
        
        slots = self.SLOTS
        limit = len(slots)
        if limit:
            index = 0
            while True:
                name = slots[index]
                index += 1
                if (name == 'handler') or name.startswith('_'):
                    if index == limit:
                        break
                    else:
                        continue
                
                # case of `channel_id`, `guild_id`
                if name.endswith('id'):
                    display_name = name[:-3]
                # case of `channel_ids`, `guild_ids`
                elif name.endswith('ids'):
                    display_name = name[:-4]
                else:
                    display_name = name
                
                result.append(display_name)
                result.append('=')
                attr = getattr(self,name)
                result.append(repr(attr))
                
                if index == limit:
                    break
                
                result.append(', ')
                continue
        
        handler = self.handler
        if (handler is not None):
            if limit:
                result.append(', ')
            result.append('handler=')
            result.append(repr(handler))
        
        result.append(')')
        
        return ''.join(result)
    
    if NEEDS_DUMMY_INIT:
        def __init__(self, *args, **kwargs):
            pass
    
    def __invert__(self):
        """Inverts the check's condition returning a new check."""
        return _invert_op_check(self)
    
    def __or__(self, other):
        """Connects the two check with `or` relation."""
        if not isinstance(other, _check_base):
            return NotImplemented
        
        return _or_op_check(self, other)

    def __and__(self, other):
        """Connects the two check with `and` relation."""
        if not isinstance(other, _check_base):
            return NotImplemented
        
        return _and_op_check(self, other)

class _invert_op_check(_check_base):
    """
    Inverts an internal check.
    
    Attributes
    ----------
    check : ``_check_base`` instance
        The inverted check.
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ('check',)
    
    def __new__(cls, check):
        """
        Creates a new reverted check.
        
        Parameters
        ----------
        check : ``_check_base`` instance
            The check to invert.
        
        Returns
        -------
        self : ``_check_base`` instance
            The reverted check.
            
            If the check is already inverted, reverts it and returns the original one.
            
            If the check has an inverted check pair, returns an instead of that instead.
        """
        if isinstance(check, cls):
            self = check.check
        else:
            target_type = CHECK_INVERT_TABLE.get(check.__class__, None)
            if target_type is None:
                self = object.__new__(cls)
                self.check = check
                self.handler = check.handler
            else:
                # None of these classes have other attributes yet, so we need just to assign their `.handler`.
                self = object.__new__(target_type)
                self.handler = check.handler
        
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        return not await self.check(client, message)


class _or_op_check(_check_base):
    """
    Creates an `or` relations between 2 checks.
    
    Attributes
    ----------
    checks : `tuple` or ``_check_base`` instances
        The or-ed checks.
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ('checks', )
    def __new__(cls, check_1, check_2):
        """
        Creates a new or-ed check.
        
        Parameters
        ----------
        check_1 : ``_check_base`` instance
            The first check to connect.
        check_2 : ``_check_base`` instance
            The second check to connect.
        
        Returns
        -------
        self : ``_check_base`` instance
            The or-ed check.
            
            If the check has natural or relation with an other check, will return that one instead.
        """
        check_1_handler = check_1.handler
        check_2_handler = check_2.handler
        if check_2_handler is None:
            handler = check_1_handler
        else:
            handler = check_2_handler
        
        check_1_type = check_1.__class__
        check_2_type = check_2.__class__
        
        
        if (check_1_type is _check_base) or (check_2_type is _check_base):
            if check_1_type is _check_base:
                if check_2_type is _check_base:
                    return _check_base(handler=handler)
                
                other_check = check_2
            
            else:
                other_check = check_1
            
            self = object.__new__(other_check.__class__)
            self.handler = handler
            
            for slot_name in other_check.SLOTS:
                if slot_name == 'handler':
                    continue
                
                slot_value = getattr(other_check, slot_name)
                setattr(self, slot_name, slot_value)
            
            return self
        
        # Permission case?
        if (check_1_type is check_2_type) and (check_1_type in CHECK_PERMISSION_TABLE):
            self = object.__new__(check_1_type)
            self.handler = handler
            self.permissions = check_1.permissions | check_2.permissions
            return self
        
        check_type = CHECK_OR_TABLE.get(frozenset((check_1_type, check_2_type)))
        if (check_type is not None):
            # Guild only cases? -> facepalm?
            if issubclass(check_1_type, guild_only) or issubclass(check_2_type, guild_only):
                self = object.__new__(check_type)
                self.handler = handler
                return self
            
            # Connectible cases?
            if issubclass(check_1_type, CHECKS_OR_ONLY_TYPES) or issubclass(check_2_type, CHECKS_OR_ONLY_TYPES):
                if issubclass(check_1_type, CHECKS_OR_ONLY_TYPES):
                    other_check = check_2
                else:
                    other_check = check_1
                
                self = object.__new__(check_type)
                self.handler = handler
                for slot_name in other_check.SLOTS:
                    if slot_name == 'handler':
                        continue
                    
                    slot_value = getattr(other_check, slot_name)
                    setattr(self, slot_name, slot_value)
                
                return self
            
            # Entity or?
            if issubclass(check_1_type, CHECKS_ENTITY_TYPES) and issubclass(check_2_type, CHECKS_ENTITY_TYPES):
                entities = []
                
                for check in (check_1, check_2):
                    slot_name = CHECKS_ENTITY_SLOTS[check.__class__]
                    slot_value = getattr(check, slot_name)
                    if isinstance(slot_value, set):
                        entities.extend(slot_value)
                    else:
                        entities.append(slot_value)
                
                self = check_type(entities, handler=handler)
                return self
            
            # Other case Should not happen
        
        checks_unwrapped = []
        if issubclass(check_1_type, cls):
            checks_unwrapped.extend(check_1.checks)
        else:
            checks_unwrapped.append(check_1)
        
        if issubclass(check_2_type, cls):
            checks_unwrapped.extend(check_2.checks)
        else:
            checks_unwrapped.append(check_2)
        
        checks = tuple(checks_unwrapped)
        
        self = object.__new__(cls)
        self.checks = checks
        self.handler = handler
        
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        for check in self.checks:
            if await check(client, message):
                return True
        
        return False

class _and_op_check(_check_base):
    """
    Creates an `or` relations between 2 checks.
    
    Attributes
    ----------
    checks : `tuple` or ``_check_base`` instances
        The or-ed checks.
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ('checks', )
    def __new__(cls, check_1, check_2):
        """
        Creates a new and-ed check.
        
        Parameters
        ----------
        check_1 : ``_check_base`` instance
            The first check to connect.
        check_2 : ``_check_base`` instance
            The second check to connect.
        
        Returns
        -------
        self : ``_check_base`` instance
            The and-ed check.
            
            If the check has natural and relation with an other check, will return that one instead.
        
        Raises
        ------
        ValueError
            The two check's have different `handler`.
        """
        check_1_handler = check_1.handler
        check_2_handler = check_2.handler
        if check_1_handler != check_2_handler:
            raise ValueError(f'The two checks have different handlers: check_1={check_1!r}, check_2={check_2!r}. '
                'Please make sure they have the same when creating `and` connection between them.')
        
        handler = check_1_handler
        
        check_1_type = check_1.__class__
        check_2_type = check_2.__class__
        
        # Empty check?
        if (check_1_type is _check_base) or (check_2_type is _check_base):
            return _check_base(handler=handler)
        
        # Permission case?
        if (check_1_type is check_2_type) and (check_1_type in CHECK_PERMISSION_TABLE):
            self = object.__new__(check_1_type)
            self.handler = handler
            self.permissions = check_1.permissions & check_2.permissions
            return self
        
        check_type = CHECK_AND_TABLE.get(frozenset((check_1_type, check_2_type)))
        if (check_type is not None):
            # Connectible cases?
            if issubclass(check_1_type, CHECKS_AND_ONLY_TYPES) or issubclass(check_2_type, CHECKS_AND_ONLY_TYPES):
                if issubclass(check_1_type, CHECKS_AND_ONLY_TYPES):
                    other_check = check_2
                else:
                    other_check = check_1
                
                self = object.__new__(check_type)
                self.handler = handler
                for slot_name in other_check.SLOTS:
                    if slot_name == 'handler':
                        continue
                    
                    slot_value = getattr(other_check, slot_name)
                    setattr(self, slot_name, slot_value)
                
                return self
            
            # Entity or?
            if issubclass(check_1_type, CHECKS_ENTITY_TYPES) and issubclass(check_2_type, CHECKS_ENTITY_TYPES):
                entities = []
                
                slot_name = CHECKS_ENTITY_SLOTS[check_1_type]
                slot_value = getattr(check_1, slot_name)
                
                if isinstance(slot_value, set):
                    entities.extend(slot_value)
                else:
                    entities.append(slot_value)
                
                slot_name = CHECKS_ENTITY_SLOTS[check_2_type]
                slot_value = getattr(check_2, slot_name)
                
                if isinstance(slot_value, set):
                    for entity in slot_value:
                        try:
                            entities.remove(entity)
                        except ValueError:
                            pass
                else:
                    try:
                        entities.remove(slot_value)
                    except ValueError:
                        pass
                
                self = check_type(entities, handler=handler)
                return self
            
            # Other case Should not happen
        
        checks_unwrapped = []
        if isinstance(check_1, cls):
            checks_unwrapped.extend(check_1.checks)
        else:
            checks_unwrapped.append(check_1)
        
        if isinstance(check_2, cls):
            checks_unwrapped.extend(check_2.checks)
        else:
            checks_unwrapped.append(check_2)
        
        checks = tuple(checks_unwrapped)
        
        self = object.__new__(cls)
        self.checks = checks
        self.handler = handler
        
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        for check in self.checks:
            if not await check(client, message):
                return False
        
        return True


class has_role(_check_base):
    """
    Checks whether a message's author has the given role.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    role : ``Role``
        The role, what the user should have.
    """
    __slots__ = ('role', )
    def __new__(cls, role, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        role : `str`, `int` or ``Role``
            The role what the message's author should have.
        handler : `None` or `async-callable` or instantiable to `async-callable`
            Will be called when the check fails.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - If `role` was not given neither as ``Role``, `str` or `int` instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        ValueError
            If `role` was given as `str` or as `int` instance, but not as a valid snowflake, so ``Role``
                instance cannot be precreated with it.
        """
        role = instance_or_id_to_instance(role, Role, 'role')
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.role = role
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        if message.author.has_role(self.role):
            return True
        
        return False


class owner_or_has_role(has_role):
    """
    Checks whether a message's author has the given role, or if it the client's owner.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    role : ``Role``
        The role, what the user should have.
    """
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        user = message.author
        if user.has_role(self.role):
            return True
        
        if client.is_owner(user):
            return True
        
        return False


class has_any_role(_check_base):
    """
    Checks whether a message's author has any of the given roles.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    roles : `set` of ``Role``
        The roles from what the user should have at least 1.
    """
    __slots__ = ('roles', )
    def __new__(cls, roles, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        roles : `iterable` of (`str`, `int` or ``Role``)
            Role from what the message's author should have at least 1.
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - If `roles` was not given as an `iterable`.
            - If an element of `roles` was not given neither as ``Role``, `str` or `int` instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        ValueError
            If an element of `roles` was given as `str` or as `int` instance, but not as a valid snowflake, so
                ``Role`` instance cannot be precreated with it.
        """
        roles_type = roles.__class__
        if not hasattr(roles_type,'__iter__'):
            raise TypeError(f'`roles` can be given as `iterable` of (`str`, `int` or `{Role.__name__}`), got '
                f'{roles_type.__name__}.')
        
        roles_processed = set()
        for role in roles:
            role = instance_or_id_to_instance(role, Role, 'role')
            roles_processed.add(role)
        
        roles_processed_length = len(roles_processed)
        if roles_processed_length == 0:
            return _check_base(handler)
        elif roles_processed_length == 1:
            return has_role(roles_processed.pop(), handler)
        
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.roles = roles_processed
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        user = message.author
        for role in self.roles:
            if  user.has_role(role):
                return True
        
        return False


class owner_or_has_any_role(has_any_role):
    """
    Checks whether a message's author has any of the given roles or if it is the client's owner.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    roles : `set` of ``Role``
        The roles from what the user should have at least 1.
    """
    def __new__(cls, roles, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        roles : `iterable` of (`str`, `int` or ``Role``)
            Role from what the message's author should have at least 1.
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - If `roles` was not given as an `iterable`.
            - If an element of `roles` was not given neither as ``Role``, `str` or `int` instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        ValueError
            If an element of `roles` was given as `str` or as `int` instance, but not as a valid snowflake, so
                ``Role`` instance cannot be precreated with it.
        """
        roles_type = roles.__class__
        if not hasattr(roles_type,'__iter__'):
            raise TypeError(f'`roles` can be given as `iterable` of (`str`, `int` or `{Role.__name__}`), got '
                f'{roles_type.__name__}.')
        
        roles_processed = set()
        for role in roles:
            role = instance_or_id_to_instance(role, Role, 'role')
            roles_processed.add(role)
        
        if len(roles_processed) == 1:
            return owner_or_has_role(roles_processed.pop(), handler)
        
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.roles = roles_processed
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        user = message.author
        for role in self.roles:
            if user.has_role(role):
                return True
        
        if client.is_owner(user):
            return True
        
        return False


class guild_only(_check_base):
    """
    Checks whether a message was sent to a guild channel.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        if (message.guild is not None):
            return True
        
        return False


class private_only(_check_base):
    """
    Checks whether a message was sent to a private channel.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        if (message.guild is None):
            return True
        
        return False

class owner_only(_check_base):
    """
    Checks whether a message was sent by the client's owner.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        if client.is_owner(message.author):
            return True
        
        return False


class guild_owner(_check_base):
    """
    Checks whether a message was sent by the message's guild's owner.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        guild = message.guild
        if guild is None:
            return False
        
        if guild.owner_id == message.author.id:
            return True
        
        return False


class owner_or_guild_owner(guild_owner):
    """
    Checks whether a message was sent by the message's guild's owner or by the client's owner.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        guild = message.guild
        if guild is None:
            return False
        
        user = message.author
        if guild.owner_id == user.id:
            return True
        
        if client.is_owner(user):
            return True
        
        return False


class has_permissions(_check_base):
    """
    Checks whether a message's author has the given permissions at the message's channel.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    permissions : ``Permission``
        The permission what the message's author should have at message's channel.
    """
    __slots__ = ('permissions', )
    def __new__(cls, permissions, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        permissions : ``Permission`` or `int` instance
            The permission, what the message's author should have at the message's channel.
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - `permissions` was not given as `int` instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        """
        permissions = _convert_permissions(permissions)
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.permissions = permissions
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        if message.channel.permissions_for(message.author) >= self.permissions:
            return True
        
        return False


class owner_or_has_permissions(has_permissions):
    """
    Checks whether a message's author has the given permissions at the message's channel, or if it is the client's
    owner.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    permissions : ``Permission``
        The permission what the message's author should have at message's channel.
    """
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        user = message.author
        if message.channel.permissions_for(user) >= self.permissions:
            return True
        
        if client.is_owner(user):
            return True
        
        return False


class has_guild_permissions(_check_base):
    """
    Checks whether a message's author has the given permissions at the message's guild.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    permissions : ``Permission``
        The permission what the message's author should have at message's guild.
    """
    __slots__ = ('permissions', )
    def __new__(cls, permissions, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        permissions : ``Permission`` or `int` instance
            The permission, what the message's author should have at the message's guild.
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - `permissions` was not given as `int` instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        """
        permissions = _convert_permissions(permissions)
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.permissions = permissions
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        guild = message.guild
        if guild is None:
            return False
        
        if guild.permissions_for(message.author) >= self.permissions:
            return True
        
        return False


class owner_or_has_guild_permissions(has_permissions):
    """
    Checks whether a message's author has the given permissions at the message's guild, or if it is the client's
    owner.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    permissions : ``Permission``
        The permission what the message's author should have at message's guild.
    """
    __slots__ = ('permissions', )
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        guild = message.guild
        if guild is None:
            return False
        
        user = message.author
        
        if guild.permissions_for(user) >= self.permissions:
            return True
        
        if client.is_owner(user):
            return True
        
        return False


class client_has_permissions(_check_base):
    """
    Checks whether a client has the given permissions at the message's channel.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    permissions : ``Permission``
        The permission what the client should have at message's channel.
    """
    __slots__ = ('permissions', )
    def __new__(cls, permissions, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        permissions : ``Permission`` or `int` instance
            The permission, what the client should have at the message's channel.
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - `permissions` was not given as `int` instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        """
        permissions = _convert_permissions(permissions)
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.permissions = permissions
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        if message.channel.cached_permissions_for(client) >= self.permissions:
            return True
        
        return False


class client_has_guild_permissions(_check_base):
    """
    Checks whether a client has the given permissions at the message's guild.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    permissions : ``Permission``
        The permission what the client should have at message's guild.
    """
    __slots__ = ('permissions', )
    def __new__(cls, permissions, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        permissions : ``Permission`` or `int` instance
            The permission, what the client should have at the message's guild.
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - `permissions` was not given as `int` instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        """
        permissions = _convert_permissions(permissions)
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.permissions = permissions
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        guild = message.guild
        if guild is None:
            return False
        
        if guild.cached_permissions_for(client) >= self.permissions:
            return True
        
        return False


class is_guild(_check_base):
    """
    Checks whether the message was sent to the given guild.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    guild_id : `int`
        The respective guild's id.
    """
    __slots__ = ('guild_id', )
    def __new__(cls, guild, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        guild : `str`, `int` or ``Guild``
            The guild where the message should be sent.
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild``, `str` or `int` instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        ValueError
            If `guild` was given as `str` or as `int` instance, but not as a valid snowflake.
        """
        guild_id = instance_or_id_to_snowflake(guild, Guild, 'guild')
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.guild_id = guild_id
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        guild = message.guild
        if guild is None:
            return False
        
        if (guild.id == self.guild_id):
            return True
        
        return False


class is_any_guild(_check_base):
    """
    Checks whether the message was sent to any of the given guilds.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    guild_ids : `set of `int`
        The respective guilds' ids.
    """
    __slots__ = ('guild_ids', )
    def __new__(cls, guilds, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        guilds : `iterable` of (`str`, `int` or ``Guild``)
            Guilds to where the message should be sent.
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - If `guilds` was not given as an `iterable`.
            - If an element of `guilds` was not given neither as ``Guild``, `str` or `int` instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        ValueError
            If an element of `guilds` was given as `str` or as `int` instance, but not as a valid snowflake.
        """
        guild_type = guilds.__class__
        if not hasattr(guild_type, '__iter__'):
            raise TypeError(f'`guilds` can be given as `iterable` of (`str`, `int` or `{Guild.__name__}`), got '
                f'{guild_type.__name__}.')
        
        guild_ids_processed = set()
        for guild in guilds:
            guild_id = instance_or_id_to_snowflake(guild, Guild, 'guild')
            guild_ids_processed.add(guild_id)
        
        guild_ids_processed_length = len(guild_ids_processed)
        if guild_ids_processed_length == 0:
            return _check_base(handler)
        elif guild_ids_processed_length == 1:
            return is_guild(guild_ids_processed.pop(), handler)
        
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.guild_ids = guild_ids_processed
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        guild = message.guild
        if guild is None:
            return False
        
        if (guild.id in self.guild_ids):
            return True
        
        return False


class custom(_check_base):
    """
    Checks whether the message and client passes the given custom condition.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    _is_async : `bool`
        Whether ``.function`` is async.
    function : `callable`
        The custom check's function.
    """
    __slots__ = ('_is_async', 'function')
    def __new__(cls, function, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        function : `callable`
            The custom check what should pass. Can be async.
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - If `function` was not given as an `callable`.
            - `function` accepts more or less non reserved positional non default parameters.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        
        Notes
        -----
        Only `int` instances are evaluated to boolean.
        """
        analyzer = CallableAnalyzer(function)
        non_reserved_positional_parameter_count = analyzer.get_non_reserved_positional_parameter_count()
        if  non_reserved_positional_parameter_count != 2:
            raise TypeError(f'The passed function: {function!r} should have accept `2` non reserved, positional, '
                f'non default parameters, meanwhile it accepts `{non_reserved_positional_parameter_count}`.')
        
        is_async = analyzer.is_async()
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.function = function
        self._is_async = is_async
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        try:
            result = self.function(client, message)
            if self._is_async:
                result = await result
        except BaseException as err:
            Task(client.events.error(client, repr(self), err), KOKORO)
            return False
        
        if result is None:
            return False
        
        if isinstance(result, int) and result:
            return True
        
        return False


class is_channel(_check_base):
    """
    Checks whether the message was sent to the given channel.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    channel_id : `int`
        The respective channel's id.
    """
    __slots__ = ('channel_id', )
    def __new__(cls, channel, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        channel : `str`, `int` or ``ChannelBase``
            The channel where the message should be sent.
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - If `channel` was not given neither as ``ChannelBase``, `str` or `int` instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        ValueError
            If `channel` was given as `str` or as `int` instance, but not as a valid snowflake.
        """
        channel_id = instance_or_id_to_snowflake(channel, ChannelBase, 'channel')
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.channel_id = channel_id
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        if (message.channel.id == self.channel_id):
            return True
        
        return False


class is_any_channel(_check_base):
    """
    Checks whether the message was sent to any of the given channels.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    channel_ids : `set` of `int`
        The respective channels' ids.
    """
    __slots__ = ('channel_ids', )
    def __new__(cls, channels, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        channels : `iterable` of (`str`, `int` or ``ChannelBase``)
            Channels to where the message should be sent.
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - If `channels` was not given as an `iterable`.
            - If an element of `channels` was not given neither as ``ChannelBase``, `str` or `int` instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        ValueError
            If an element of `channels` was given as `str` or as `int` instance, but not as a valid snowflake.
        """
        channels_type = channels.__class__
        if not hasattr(channels_type, '__iter__'):
            raise TypeError(f'`channels` can be given as `iterable` of (`str`, `int` or `{ChannelBase.__name__}`), '
                f'got {channels_type.__name__}.')
        
        channel_ids_processed = set()
        for channel in channels:
            channel_id = instance_or_id_to_snowflake(channel, ChannelBase, 'channel')
            channel_ids_processed.add(channel_id)
        
        channel_ids_processed_length = len(channel_ids_processed)
        if channel_ids_processed_length == 0:
            return _check_base(handler)
        elif channel_ids_processed_length == 1:
            return is_channel(channel_ids_processed.pop(), handler)
        
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.channel_ids = channel_ids_processed
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        if (message.channel.id in self.channel_ids):
            return True
        
        return False


class nsfw_channel_only(_check_base):
    """
    Checks whether the message was sent to an nsfw channel.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        channel = message.channel
        if (isinstance(channel, ChannelText) and channel.nsfw):
            return True
        
        return False

class announcement_channel_only(_check_base):
    """
    Checks whether the message was sent to an announcement channel.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        channel = message.channel
        if channel.type == 5:
            return True
        
        return False


class is_in_voice(_check_base):
    """
    Checks whether the message was sent by a user who is in any of the respective guild's voice channel.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        guild = message.guild
        if guild is None:
            return False
        
        if message.author.id in guild.voice_states:
            return True
        
        return False


class booster_only(_check_base):
    """
    Checks whether the message was sent by a user who boosts the respective server.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        guild = message.guild
        if guild is None:
            return
        
        try:
            profile = message.author.guild_profiles[guild.id]
        except KeyError:
            return False
        
        if profile.boosts_since is None:
            return False
        
        return True


class client_only(_check_base):
    """
    Checks whether the message was sent by a client.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        if isinstance(message.author, Client):
            return True
        
        return False


class user_account_only(_check_base):
    """
    Checks whether the message was sent by a user account.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        if message.author.is_bot:
            return False
        
        return True


class bot_account_only(_check_base):
    """
    Checks whether the message was sent by a bot account.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        if message.author.is_bot:
            return True
        
        return False


class user_account_or_client(_check_base):
    """
    Checks whether the message was sent by a user account or by a ``Client`` instance.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    """
    __slots__ = ()
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        user = message.author
        if not user.is_bot:
            return True
        
        if isinstance(user, Client):
            return True
        
        return False


class is_in_category(_check_base):
    """
    Checks whether the message was sent to a channel within the given category.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    category_id : `int`
        The respective category's id.
    """
    __slots__ = ('category_id', )
    def __new__(cls, category, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        category : `str`, `int`, ``ChannelCategory`` or ``Guild``
            The category, within sent messages pass the check.
            
            If you want to check those channels, which are not in any category, pass the respective ``Guild`` instead.
        
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - If `category` was not given neither as ``ChannelCategory``, ``Guild``, `str` or `int` instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        ValueError
            If `category` was given as `str` or as `int` instance, but not as a valid snowflake.
        """
        category_id = instance_or_id_to_snowflake(category, (ChannelCategory, Guild), 'category')
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.category_id = category_id
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        channel = message.channel
        if not isinstance(channel, ChannelGuildBase):
            return False
        
        parent = channel.parent
        if parent is None:
            guild = channel.guild
            if guild is None:
                return False
            
            parent_id = guild.id
        else:
            parent_id = parent.id
        
        if parent_id == self.category_id:
            return True
        
        return False


class is_in_any_category(_check_base):
    """
    Checks whether the message was sent to a channel within any of the given category.
    
    Attributes
    ----------
    handler : `None` or `async-callable`
        An async callable what will be called when the check fails.
    category_ids : `set` of `int`
        The respective category's id.
    """
    __slots__ = ('category_ids', )
    def __new__(cls, categories, handler=None):
        """
        Creates a check, what will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        categories : `iterable` of (`str`, `int`, ``ChannelCategory`` or ``Guild``)
            The categories, within sent messages pass the check.
            
            If you want to check those channels, which are not in any category, pass the respective ``Guild`` instead.
        
        handler : `None` or `async-callable` or instantiable to `async-callable`
            The handler to convert.
            
            If the handler is `async-callable` or if it would be instanced to it, then it should accept the
            following parameters:
            +-------------------+---------------------------+
            | Respective name   | Type                      |
            +===================+===========================+
            | client            | ``Client``                |
            +-------------------+---------------------------+
            | message           | ``Message``               |
            +-------------------+---------------------------+
            | command           | ``Command`` or `str`      |
            +-------------------+---------------------------+
            | check             | ``_check_base`` instance  |
            +-------------------+---------------------------+
        
        Raises
        ------
        TypeError
            - If `categories` was not given as an `iterable`.
            - If an element of `categories` was not given neither as ``ChannelCategory``, ``Guild``, `str` or `int`
                instance.
            - If `handler` was given as an invalid type, or it accepts a bad amount of parameters.
        ValueError
            If an element of `categories` was given as `str` or as `int` instance, but not as a valid snowflake.
        """
        categories_type = categories.__class__
        if not hasattr(categories_type, '__iter__'):
            raise TypeError(f'`categories` can be given as `iterable` of (`str`, `int`, `{ChannelCategory.__name__}`, '
                f'or `{Guild.__name__}`), got {categories_type.__name__}.')
        
        category_ids_processed = set()
        for category in categories:
            category_id = instance_or_id_to_snowflake(category, (ChannelCategory, Guild), 'category')
            category_ids_processed.add(category_id)
        
        category_ids_processed_length = len(category_ids_processed)
        if category_ids_processed_length == 0:
            return _check_base(handler)
        elif category_ids_processed_length == 1:
            return is_in_category(category_ids_processed.pop(), handler)
        
        handler = _convert_handler(handler)
        
        self = object.__new__(cls)
        self.category_ids = category_ids_processed
        self.handler = handler
        return self
    
    async def __call__(self, client, message):
        """
        Calls the check to validate whether it passes with the given conditions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who's received the message.
        message : ``Message``
            The received message.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
        channel = message.channel
        if not isinstance(channel, ChannelGuildBase):
            return False
        
        parent = channel.parent
        if parent is None:
            guild = channel.guild
            if guild is None:
                return False
            
            parent_id = guild.id
        else:
            parent_id = parent.id
        
        if parent_id in self.category_ids:
            return True
        
        return False


CHECK_INVERT_TABLE = {
    bot_account_only: user_account_only,
    guild_only: private_only,
    private_only: guild_only,
    user_account_only: bot_account_only,
}

CHECKS_ENTITY_SLOTS = {
    is_channel: 'channel_id',
    is_guild: 'guild_id',
    is_in_category: 'category_id',
    has_role: 'role',
    owner_or_has_role: 'role',
    is_any_channel: 'channel_ids',
    is_any_guild: 'guild_ids',
    is_in_any_category: 'category_ids',
    has_any_role: 'roles',
    owner_or_has_any_role: 'roles',
}

CHECKS_ENTITY_TYPES = (
    is_channel,
    is_guild,
    is_in_category,
    has_role,
    owner_or_has_role,
    is_any_channel,
    is_any_guild,
    is_in_any_category,
    has_any_role,
    owner_or_has_any_role,
)

CHECKS_OR_ONLY_TYPES = (
    owner_only,
    client_only,
)

CHECK_OR_TABLE = {
    frozenset((guild_only, client_has_guild_permissions)): guild_only,
    frozenset((guild_only, booster_only)): guild_only,
    frozenset((guild_only, has_guild_permissions)): guild_only,
    frozenset((guild_only, guild_only)): guild_only,
    frozenset((guild_only, guild_owner)): guild_only,
    frozenset((guild_only, has_guild_permissions)): guild_only,
    frozenset((guild_only, is_guild)): guild_only,
    frozenset((guild_only, is_any_guild)): guild_only,
    frozenset((guild_only, is_in_category)): guild_only,
    frozenset((guild_only, is_in_any_category)): guild_only,
    frozenset((guild_only, owner_or_guild_owner)): guild_only,
    frozenset((guild_only, owner_or_has_guild_permissions)): guild_only,
    
#   Note, that `guild_owner` and `has_guild_permissions` require to be inside of a `guild`, so we cannot take them.
#   frozenset((owner_only, guild_owner)): owner_or_guild_owner,
#   frozenset((owner_only, has_guild_permissions)): owner_or_has_guild_permissions,
    
    frozenset((owner_only, has_any_role)): owner_or_has_any_role,
    frozenset((owner_only, has_permissions)): owner_or_has_permissions,
    frozenset((owner_only, has_role)): owner_or_has_role,
    
    frozenset((client_only, user_account_only)): user_account_or_client,
    
    frozenset((is_channel, is_channel)): is_any_channel,
    frozenset((is_channel, is_any_channel)): is_any_channel,
    frozenset((is_any_channel, is_any_channel)): is_any_channel,
    frozenset((is_guild, is_guild)): is_any_guild,
    frozenset((is_guild, is_any_guild)): is_any_guild,
    frozenset((is_any_guild, is_any_guild)): is_any_guild,
    frozenset((is_in_category, is_in_any_category)): is_in_any_category,
    frozenset((is_in_category, is_in_any_category)): is_in_any_category,
    frozenset((is_in_any_category, is_in_any_category)): is_in_any_category,
    frozenset((has_role, has_role)): has_any_role,
    frozenset((has_role, has_any_role)): has_any_role,
    frozenset((has_any_role, has_any_role)): has_any_role,
    frozenset((owner_or_has_role, owner_or_has_role)): owner_or_has_any_role,
    frozenset((owner_or_has_role, owner_or_has_any_role)): owner_or_has_any_role,
    frozenset((owner_or_has_any_role, owner_or_has_any_role)): owner_or_has_any_role,
}


CHECK_PERMISSION_TABLE = {
    client_has_guild_permissions,
    client_has_permissions,
    has_guild_permissions,
    has_permissions,
    owner_or_has_guild_permissions,
    owner_or_has_permissions,
}

CHECKS_AND_ONLY_TYPES = (
    guild_only,
)

CHECK_AND_TABLE = {
    frozenset((guild_only, client_has_guild_permissions)): client_has_guild_permissions,
    frozenset((guild_only, booster_only)): booster_only,
    frozenset((guild_only, has_guild_permissions)): has_guild_permissions,
    frozenset((guild_only, guild_only)): guild_only,
    frozenset((guild_only, guild_owner)): guild_owner,
    frozenset((guild_only, has_guild_permissions)): has_guild_permissions,
    frozenset((guild_only, is_guild)): is_guild,
    frozenset((guild_only, is_any_guild)): is_any_guild,
    frozenset((guild_only, is_in_category)): is_in_category,
    frozenset((guild_only, is_in_any_category)): is_in_any_category,
    frozenset((guild_only, owner_or_guild_owner)): owner_or_guild_owner,
    frozenset((guild_only, owner_or_has_guild_permissions)): owner_or_has_guild_permissions,
    
    frozenset((is_channel, is_channel)): is_any_channel,
    frozenset((is_channel, is_any_channel)): is_any_channel,
    frozenset((is_any_channel, is_any_channel)): is_any_channel,
    frozenset((is_guild, is_guild)): is_any_guild,
    frozenset((is_guild, is_any_guild)): is_any_guild,
    frozenset((is_any_guild, is_any_guild)): is_any_guild,
    frozenset((is_in_category, is_in_any_category)): is_in_any_category,
    frozenset((is_in_category, is_in_any_category)): is_in_any_category,
    frozenset((is_in_any_category, is_in_any_category)): is_in_any_category,
    frozenset((has_role, has_role)): has_any_role,
    frozenset((has_role, has_any_role)): has_any_role,
    frozenset((has_any_role, has_any_role)): has_any_role,
    frozenset((owner_or_has_role, owner_or_has_role)): owner_or_has_any_role,
    frozenset((owner_or_has_role, owner_or_has_any_role)): owner_or_has_any_role,
    frozenset((owner_or_has_any_role, owner_or_has_any_role)): owner_or_has_any_role,
}



import sys, warnings
from code import InteractiveConsole
from functools import partial as partial_func
from types import FunctionType
try:
    import readline
except ImportError:
    pass
from .. import Future, EventThread, is_awaitable, __package__ as PACKAGE_NAME

PACKAGE = __import__(PACKAGE_NAME)

# If only hata backend is imported, KOKORO is not present.
try:
    from .. import KOKORO as EVENT_LOOP
except ImportError:
    EVENT_LOOP = EventThread(daemon=False)

# On python3.6 not present?
try:
    from ast import PyCF_ALLOW_TOP_LEVEL_AWAIT
except ImportError:
    PyCF_ALLOW_TOP_LEVEL_AWAIT = 1<<13
    AWAIT_AVAILABLE = False
else:
    AWAIT_AVAILABLE = True


def run_code_callback(console, code):
    function = FunctionType(code, console.locals)
    try:
        coroutine = function()
    except BaseException as err:
        if isinstance(err, KeyboardInterrupt):
            console.interrupted = True
        
        console.future.set_exception(err)
        return
    
    if not is_awaitable(coroutine):
        console.future.set_result(coroutine)
        return
    
    try:
        task = EVENT_LOOP.ensure_future(coroutine)
    except BaseException as err:
        console.future.set_exception(err)
    else:
        console.future.set_result(None)
        console.task = task


class AsynchronousInteractiveConsole(InteractiveConsole):
    
    def __init__(self, locals):
        InteractiveConsole.__init__(self, locals)
        self.compile.compiler.flags |= PyCF_ALLOW_TOP_LEVEL_AWAIT
        
        self.future = None
        self.interrupted = False
        self.task = None
    
    def runcode(self, code):
        self.future = Future(EVENT_LOOP)
        self.interrupted = False
        self.task = None
        
        EVENT_LOOP.call_soon_thread_safe(partial_func(run_code_callback, self, code))
        
        try:
            result = self.future.sync_wrap().wait()
            
            task = self.task
            if (task is not None):
                result = task.sync_wrap().wait()
        
        except BaseException as err:
            future = self.future
            if (not future.done()):
                future.cancel()
            
            task = self.task
            if (task is not None) and (not task.done()):
                task.cancel()
            
            if isinstance(err, SystemExit):
                raise
            
            if self.interrupted:
                self.write('\nKeyboardInterrupt\n')
            else:
                self.showtraceback()
        else:
            return result

if AWAIT_AVAILABLE:
    AWAIT_NOTE = 'Use \'await\' directly.'
else:
    AWAIT_NOTE = '!!! Direct \'await\' is not available on your python version. Please use python 3.8 or newer !!!'

NAME = 'interpreter'
USAGE = 'i | interpreter'

HELP = (
    f'Runs asynchronous python interpreter through hata\'s asynchronous environment.\n'
    f'{AWAIT_NOTE}\n'
)

def __main__():
    interactive_console_locals = {PACKAGE_NAME: PACKAGE}
    for variable_name in {
            '__name__',
            '__package__',
            '__loader__',
            '__spec__',
            '__builtins__',
            '__file__'
                }:
        interactive_console_locals[variable_name] = getattr(PACKAGE, variable_name)
    
    for variable_name in PACKAGE.__all__:
        interactive_console_locals[variable_name] = getattr(PACKAGE, variable_name)
    

    
    banner = (
        f'{PACKAGE_NAME} interactive_console {sys.version} on {sys.platform}.\n'
        f'{AWAIT_NOTE}\n'
        f'Type \'help\', \'copyright\', \'credits\' or \'license\' for more information.'
    )
    
    exit_message = f'exiting {PACKAGE_NAME} interactive_console...'
    
    console = AsynchronousInteractiveConsole(interactive_console_locals)
    
    try:
        console.interact(
            banner = banner,
            exitmsg = exit_message,
        )
    finally:
        warnings.filterwarnings(
            'ignore',
            message = r'^coroutine .* was never awaited$',
            category = RuntimeWarning,
        )
        
        EVENT_LOOP.stop()



# -*- coding: utf-8 -*-
import sys, warnings
from types import ModuleType

from . import asyncio

try:
    module = sys.modules['asyncio']
except KeyError:
    module = asyncio
    sys.modules['asyncio'] = asyncio

else:
    warnings.warn(f'asyncio is already loaded, replacing it.')
    
    try:
        del sys.modules['_asyncio']
    except KeyError:
        pass
    
    module.__dict__.clear()
    module.__dict__.update(asyncio.__dict__)

module.__file__ = __file__

sys.modules[__name__] = asyncio

for sub_module_name, feature_names, extra_features in (
        ('base_events', (
            'BaseEventLoop',
            '_run_until_complete_cb', # Required by anyio
            ), None,
        ),
        ('base_futures', None, None),
        ('base_subprocess', None, None),
        ('base_tasks', None, None),
        ('constants', None, None),
        ('coroutines', (
            'coroutine',
            'iscoroutinefunction',
            'iscoroutine',
            ), (
            ('_DEBUG', False), # Required by aiohttp 3.8
            )
        ),
        ('events', (
            'AbstractEventLoopPolicy',
            'AbstractEventLoop',
            'AbstractServer',
            'Handle',
            'TimerHandle',
            'get_event_loop_policy',
            'set_event_loop_policy',
            'get_event_loop',
            'set_event_loop',
            'new_event_loop',
            'get_child_watcher',
            'set_child_watcher',
            '_set_running_loop',
            'get_running_loop',
            '_get_running_loop',
            ), None,
        ),
        ('exceptions', (
            'CancelledError',
            'InvalidStateError',
            'TimeoutError',
            'IncompleteReadError',
            'LimitOverrunError',
            'SendfileNotAvailableError',
            ), None,
        ),
        ('format_helpers', None, None),
        ('futures', (
            'Future',
            'wrap_future',
            'isfuture',
            ), None,
        ),
        ('locks', (
            'Lock',
            'Event',
            'Condition',
            'Semaphore',
            'BoundedSemaphore',
            ), None,
        ),
        ('proactor_events', (
            'BaseProactorEventLoop',
            ), None,
        ),
        ('protocols', (
            'BaseProtocol',
            'Protocol',
            'DatagramProtocol',
            'SubprocessProtocol',
            'BufferedProtocol',
            ), None,
        ),
        ('queues', (
            'Queue',
            'PriorityQueue',
            'LifoQueue',
            'QueueFull',
            'QueueEmpty',
            ), None,
        ),
        ('runners', (
            'run',
            ), None,
        ),
        ('selector_events', (
            'BaseSelectorEventLoop',
            ), None,
        ),
        ('sslproto', None, None),
        ('staggered_race', (
            'staggered_race',
            ), None,
        ),
        ('streams', (
            'StreamReader',
            'StreamWriter',
            'StreamReaderProtocol',
            'open_connection',
            'start_server',
            ), None,
        ),
        ('subprocess', (
            'create_subprocess_exec',
            'create_subprocess_shell',
            'Process', # Required by anyio
            ), None,
        ),
        ('tasks', (
            'Task',
            'create_task',
            'FIRST_COMPLETED',
            'FIRST_EXCEPTION',
            'ALL_COMPLETED',
            'wait',
            'wait_for',
            'as_completed',
            'sleep',
            'gather',
            'shield',
            'ensure_future',
            'run_coroutine_threadsafe',
            'current_task',
            'all_tasks',
            '_register_task',
            '_unregister_task',
            '_enter_task',
            '_leave_task',
            ), None,
        ),
        ('threads', (
            'to_thread',
            ), None,
        ),
        ('transports', (
            'BaseTransport',
            'ReadTransport',
            'WriteTransport',
            'Transport',
            'DatagramTransport',
            'SubprocessTransport',
            ), None,
        ),
        ('trsock', None, None),
        ('unix_events', (
            'SelectorEventLoop',
            'AbstractChildWatcher',
            'SafeChildWatcher',
            'FastChildWatcher',
            'PidfdChildWatcher',
            'MultiLoopChildWatcher',
            'ThreadedChildWatcher',
            'DefaultEventLoopPolicy',
            ), None,
        ),
        ('windows_events', (
            'SelectorEventLoop',
            'ProactorEventLoop',
            'IocpProactor',
            'DefaultEventLoopPolicy',
            'WindowsSelectorEventLoopPolicy',
            'WindowsProactorEventLoopPolicy',
            ), None,
        ),
        ('windows_utils', (
            'pipe',
            'Popen',
            'PIPE',
            'PipeHandle',
            ), None,
        )
            ):
    
    module_name = f'asyncio.{sub_module_name}'
    
    try:
        module = sys.modules[module_name]
    except KeyError:
        module = ModuleType(module_name)
        sys.modules[module_name] = module
    else:
        module.__dict__.clear()
    
    asyncio.__dict__[sub_module_name] = module
    module.__file__ = __file__
    
    if (feature_names is not None):
        module.__all__ = feature_names
        module.__dict__.update((feature_name, getattr(asyncio, feature_name)) for feature_name in feature_names)
    
    if (extra_features is not None):
        module.__dict__.update(extra_features)

del extra_features
del sub_module_name
del feature_names
del module_name
del module
del asyncio

from .. import register_library_extension
register_library_extension('HuyaneMatsu.asyncio')
del register_library_extension



import warnings

from ...backend.export import export

from .activity_base import *
from .activity_custom import *
from .activity_rich import *
from .activity_types import *
from .activity_unknown import *
from .flags import *
from .utils import *

from . import activity_types as ACTIVITY_TYPES

export(ACTIVITY_TYPES, 'ACTIVITY_TYPES')

__all__ = (
    'ACTIVITY_TYPES',
    *activity_base.__all__,
    *activity_custom.__all__,
    *activity_rich.__all__,
    *activity_types.__all__,
    *activity_unknown.__all__,
    *flags.__all__,
    *utils.__all__,
)
